

/*
--
-- Copyright (C) 2014  <fastrgv@gmail.com>
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You may read the full text of the GNU General Public License
-- at <http://www.gnu.org/licenses/>.
--
*/





#include "lodepng.h"
#include <iostream>
using std::cout;
using std::endl;


#include <vector>
#include <GL/glew.h>
#include <cassert>








// WARNING:  for this we assume RGBA data
GLuint loadCubePNG(

const std::string  imagestr0,
const std::string  imagestr1,
const std::string  imagestr2,
const std::string  imagestr3,
const std::string  imagestr4,
const std::string  imagestr5

)
{
	const char * imagepath[6] = { 
		imagestr0.c_str(), imagestr1.c_str(), imagestr2.c_str(),
		imagestr3.c_str(), imagestr4.c_str(), imagestr5.c_str() };

	// Create one OpenGL texture
	GLuint CubeMap_texid;
	glGenTextures(1, &CubeMap_texid);

	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_CUBE_MAP);

	// "Bind" the newly created texture
	glBindTexture(GL_TEXTURE_CUBE_MAP, CubeMap_texid);

	unsigned error;
	unsigned width,height;
	unsigned oldw, oldh;

	for(int i=0; i<6; i++) {
		std::vector<unsigned char> image;
		error = lodepng::decode( image, width, height, imagepath[i] );
		if(error){cout<<"loadCubePNG: 0decoder error "<<error<<": "<<lodepng_error_text(error)<<endl;}
		unsigned char * data = &image[0];

		if( i == 0 )	
		{
			assert( height == width );
		}
		else
		{
			assert(  width == oldw );
			assert( height == oldh );
		}
		oldw=width; oldh=height;

		glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_RGBA, 
			width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

	} // end for i loop

	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	//glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR); ?BAD?
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	return CubeMap_texid;

} // end loadCubePNG



















GLuint loadPNG(const std::string imagestr, bool & Ok){

	Ok=true;

	const char * imagepath( imagestr.c_str() );

	// Create one OpenGL texture
	GLuint textureID;
	glGenTextures(1, &textureID);

	// "Bind" the newly created texture
	glBindTexture(GL_TEXTURE_2D, textureID);


	std::vector<unsigned char> image;
	unsigned width,height;
	unsigned error = lodepng::decode(image, width, height, imagepath);
	if(error){
		Ok=false;
		cout<<"loadPNG: decoder error "<<error<<": "<<lodepng_error_text(error)<<endl;
	}

	unsigned char * data = &image[0];


	//            target       lvl  ifmt                brdr  xfmt       type
	glTexImage2D(GL_TEXTURE_2D, 0,GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);


	// Nice trilinear filtering.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
	//glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR); 
	glGenerateMipmap(GL_TEXTURE_2D);

	// Return the ID of the texture we just created
	return textureID;

} // end loadPNG




GLuint loadPNG_RGB(const std::string imagestr){

	const char * imagepath( imagestr.c_str() );

	// Create one OpenGL texture
	GLuint textureID;
	glGenTextures(1, &textureID);

	// "Bind" the newly created texture
	glBindTexture(GL_TEXTURE_2D, textureID);


	std::vector<unsigned char> image;
	unsigned width,height;
	unsigned error = lodepng::decode(image, width, height, imagepath);
	if(error){
		cout<<"loadPNG: decoder error "<<error<<": "<<lodepng_error_text(error)<<endl;
	}

	unsigned char * data = &image[0];


	//            target       lvl  ifmt                brdr  xfmt       type
	glTexImage2D(GL_TEXTURE_2D, 0,GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, data);


	// Nice trilinear filtering.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR); 
	glGenerateMipmap(GL_TEXTURE_2D);

	// Return the ID of the texture we just created
	return textureID;

} // end loadPNG_RGB




