

/*
--
-- Copyright (C) 2014  <fastrgv@gmail.com>
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You may read the full text of the GNU General Public License
-- at <http://www.gnu.org/licenses/>.
--
*/


// sdl2 version
// 12jun14 <return> key alternative to right-mouse-btn
// Note:  works Ok on OSX-Retina display.
// Also has SFML applause.


#include <iostream>
using std::cout;
using std::endl;



// Include GLEW
#include <GL/glew.h>



#if defined(__linux__)

	#include <SDL2/SDL.h>
	#include <SDL2/SDL_opengl.h>

#elif defined(__APPLE__) && defined(__MACH__)

	#include <SDL2/SDL.h>
	#include <SDL2/SDL_opengl.h>

#endif



// Include GLM
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>

#include <shader.hpp>
//#include <controls3.hpp>


#include <text2d.hpp>




#include <SFML/Audio.hpp>
sf::Sound sound;
sf::SoundBuffer sbClap;



#include <string>
using std::string;



SDL_Window* mainWindow;
SDL_GLContext mainGLContext;



const double fourthpi( atan(1.0) );
const double halfpi( 2.0*fourthpi );
const double onepi( 4.0*fourthpi );
const double twopi( 2.0*onepi );
const double rad2deg( 180.0/onepi );
const double deg2rad( onepi/180.0 );

static glm::mat4 
	cumModelMatrix(glm::mat4(1.0)), ModelMatrix(glm::mat4(1.0)), 
	ViewMatrix, ProjectionMatrix, MVP;


void initModel(void){

   //here we rotate cube into preferred orientation:

	ModelMatrix = glm::mat4(1.0); //loadidentity
	glm::vec3 zaxis(0.0f,0.0f,1.0f);
	glm::vec3 yaxis(0.0f,1.0f,0.0f);
	glm::vec3 xaxis(1.0f,0.0f,0.0f);
	ModelMatrix = glm::rotate( ModelMatrix,   90.0f, yaxis );
	ModelMatrix = glm::rotate( ModelMatrix, -120.0f, zaxis );
	ModelMatrix = ModelMatrix * cumModelMatrix; //apply prior rotations
	cumModelMatrix = ModelMatrix; // update cumulative for next time

} // end initModel




//======== begin random =======================


#include <random>
std::default_random_engine rng;
std::uniform_real_distribution<float> unif01(0.0, 1.0);

//usage:  n = unif01(rng);



//========== end random =======================


//////// begin autosolver block /////////////////////////////////////////
char sol[100000]; // u,d,l,r,n,a (up, down, left, right, near, away)
int nsol(0);
//////// end autosolver block ///////////////////////////////////////////


//static const bool rgbskin(true);
static bool rgbskin(false);

static double lastTime;


static int const size(3);
static float const midrng(1.0);

// function inputs are zero-based (r,c,l):
int indx(int r, int c, int l){ return r*size*size+c*size+l; } //0..26

static int perm[ size*size*size ];
static int brow(1), bcol(1), blay(1); //zero-based
static bool winner(false), show_goal(true), 
				show_axes(false), hint(false), help(false);



//    bottom mag         middle            top red
static const string symbol[ size*size*size ]=
	{ "A", "B", "C",   "D", "E", "F",   "G", "H", "I",   //rear blu
	  "J", "K", "L",   "M", " ", "N",   "O", "P", "Q",   //mid
	  "R", "S", "T",   "U", "V", "W",   "X", "Y", "Z" }; //front cya
//   lef       rit    lef       rit     lef      rit
//   grn       yel    grn       yel     grn      yel

//  top = red,    bot = mag(1,0,1)
//  lef = grn,    rit = yel(1,1,0)
//  ant = blu,    pos = cya(0,1,1)


static float  vb[108*27];
static float rgb[108*27];


static float xxx[27], yyy[27], zzz[27];
static float red27[27], grn27[27], blu27[27];

static float red[6*27],grn[6*27],blu[6*27];






// left face green
void ccwNx(
	float r, float g, float b,
	float xc, float yc, float zc, float rr, float * v, float * u)
{
	const float xm(xc-rr), xp(xc+rr);
	const float ym(yc-rr), yp(yc+rr);
	const float zm(zc-rr), zp(zc+rr);

	v[ 0]=xm; v[1]=ym; v[2]=zm;
	v[ 3]=xm; v[4]=ym; v[5]=zp;
	v[ 6]=xm; v[7]=yp; v[8]=zp;

	v[ 9]=xm; v[10]=ym; v[11]=zm;
	v[12]=xm; v[13]=yp; v[14]=zp;
	v[15]=xm; v[16]=yp; v[17]=zm;

	for(int i=0; i<18; i+=3)
	{
		u[i+0]=r;
		u[i+1]=g;
		u[i+2]=b;
	}

}

// right face yellow
void ccwPx(
	float r, float g, float b,
	float xc, float yc, float zc, float rr, float * v, float * u)
{
	const float xm(xc-rr), xp(xc+rr);
	const float ym(yc-rr), yp(yc+rr);
	const float zm(zc-rr), zp(zc+rr);

	v[ 0]=xp; v[1]=ym; v[2]=zm;
	v[ 3]=xp; v[4]=yp; v[5]=zp;
	v[ 6]=xp; v[7]=ym; v[8]=zp;

	v[ 9]=xp; v[10]=ym; v[11]=zm;
	v[12]=xp; v[13]=yp; v[14]=zm;
	v[15]=xp; v[16]=yp; v[17]=zp;

	for(int i=0; i<18; i+=3)
	{
		u[i+0]=r;
		u[i+1]=g;
		u[i+2]=b;
	}

}





// bottom face magenta
void ccwNy(
	float r, float g, float b,
	float xc, float yc, float zc, float rr, float * v, float * u)
{
	const float xm(xc-rr), xp(xc+rr);
	const float ym(yc-rr), yp(yc+rr);
	const float zm(zc-rr), zp(zc+rr);

	v[ 0]=xp; v[1]=ym; v[2]=zp;
	v[ 3]=xm; v[4]=ym; v[5]=zm;
	v[ 6]=xp; v[7]=ym; v[8]=zm;

	v[ 9]=xp; v[10]=ym; v[11]=zp;
	v[12]=xm; v[13]=ym; v[14]=zp;
	v[15]=xm; v[16]=ym; v[17]=zm;

	for(int i=0; i<18; i+=3)
	{
		u[i+0]=r;
		u[i+1]=g;
		u[i+2]=b;
	}

}

// top face red
void ccwPy(
	float r, float g, float b,
	float xc, float yc, float zc, float rr, float * v, float * u)
{
	const float xm(xc-rr), xp(xc+rr);
	const float ym(yc-rr), yp(yc+rr);
	const float zm(zc-rr), zp(zc+rr);

	v[ 0]=xp; v[1]=yp; v[2]=zp;
	v[ 3]=xp; v[4]=yp; v[5]=zm;
	v[ 6]=xm; v[7]=yp; v[8]=zm;

	v[ 9]=xp; v[10]=yp; v[11]=zp;
	v[12]=xm; v[13]=yp; v[14]=zm;
	v[15]=xm; v[16]=yp; v[17]=zp;

	for(int i=0; i<18; i+=3)
	{
		u[i+0]=r;
		u[i+1]=g;
		u[i+2]=b;
	}

}






// rear face blue
void ccwNz(
	float r, float g, float b,
	float xc, float yc, float zc, float rr, float * v, float * u)
{
	const float xm(xc-rr), xp(xc+rr);
	const float ym(yc-rr), yp(yc+rr);
	const float zm(zc-rr), zp(zc+rr);

	v[ 0]=xp; v[1]=yp; v[2]=zm;
	v[ 3]=xm; v[4]=ym; v[5]=zm;
	v[ 6]=xm; v[7]=yp; v[8]=zm;

	v[ 9]=xp; v[10]=yp; v[11]=zm;
	v[12]=xp; v[13]=ym; v[14]=zm;
	v[15]=xm; v[16]=ym; v[17]=zm;

	for(int i=0; i<18; i+=3)
	{
		u[i+0]=r;
		u[i+1]=g;
		u[i+2]=b;
	}

}


// front face cyan
void ccwPz(
	float r, float g, float b,
	float xc, float yc, float zc, float rr, float * v, float * u)
{
	const float xm(xc-rr), xp(xc+rr);
	const float ym(yc-rr), yp(yc+rr);
	const float zm(zc-rr), zp(zc+rr);

	v[ 0]=xp; v[1]=yp; v[2]=zp;
	v[ 3]=xm; v[4]=yp; v[5]=zp;
	v[ 6]=xm; v[7]=ym; v[8]=zp;

	v[ 9]=xp; v[10]=yp; v[11]=zp;
	v[12]=xm; v[13]=ym; v[14]=zp;
	v[15]=xp; v[16]=ym; v[17]=zp;

	for(int i=0; i<18; i+=3)
	{
		u[i+0]=r;
		u[i+1]=g;
		u[i+2]=b;
	}

}











// function to generate 12 triangles making up the faces
// of a cube centered at (xc,yc,zc) with radius=rr.
// This is 36 vertices => 108 coords
// thus the 4th parm is declared as "float v[108]"
void cubelet0(
	int p,
	float xc, float yc, float zc, float rr,
	float * v, float * u )
{
	if( xc < 0 )
	{
		ccwNx(0,1,0,xc,yc,zc,rr, &v[ 0], &u[ 0]); //lef face green
		red[6*p+0]=0;
		grn[6*p+0]=1;
		blu[6*p+0]=0;
	}
	else
	{
		ccwNx(0,0,0,xc,yc,zc,rr, &v[ 0], &u[ 0]); //lef face green
		red[6*p+0]=0;
		grn[6*p+0]=0;
		blu[6*p+0]=0;
	}

	if( xc > 0 )
	{
		ccwPx(0.9,0.9,0,xc,yc,zc,rr, &v[18], &u[18]); //rit face yellow 
		red[6*p+1]=0.9;
		grn[6*p+1]=0.9; // 1,1,0 is too bright, use 1,1,0.3
		blu[6*p+1]=0;
	}
	else
	{
		ccwPx(0,0,0,xc,yc,zc,rr, &v[18], &u[18]); //rit face yellow
		red[6*p+1]=0;
		grn[6*p+1]=0;
		blu[6*p+1]=0;
	}


	if( yc < 0 )
	{
		ccwNy(1,0,1,xc,yc,zc,rr, &v[36], &u[36]); //bot face magenta
		red[6*p+2]=1;
		grn[6*p+2]=0;
		blu[6*p+2]=1;
	}
	else
	{
		ccwNy(0,0,0,xc,yc,zc,rr, &v[36], &u[36]); //bot face magenta
		red[6*p+2]=0;
		grn[6*p+2]=0;
		blu[6*p+2]=0;
	}

	if( yc > 0 )
	{
		ccwPy(1,0,0,xc,yc,zc,rr, &v[54], &u[54]); //top face red
		red[6*p+3]=1;
		grn[6*p+3]=0;
		blu[6*p+3]=0;
	}
	else
	{
		ccwPy(0,0,0,xc,yc,zc,rr, &v[54], &u[54]); //top face red
		red[6*p+3]=0;
		grn[6*p+3]=0;
		blu[6*p+3]=0;
	}


	if( zc < 0 )
	{
		ccwNz(0,0,1,xc,yc,zc,rr, &v[72], &u[72]); //pos face blue
		red[6*p+4]=0;
		grn[6*p+4]=0;
		blu[6*p+4]=1;
	}
	else
	{
		ccwNz(0,0,0,xc,yc,zc,rr, &v[72], &u[72]); //pos face blue
		red[6*p+4]=0;
		grn[6*p+4]=0;
		blu[6*p+4]=0;
	}

	if( zc > 0 )
	{
		ccwPz(0,1,1,xc,yc,zc,rr, &v[90], &u[90]); //ant face cyan
		red[6*p+5]=0;
		grn[6*p+5]=1;
		blu[6*p+5]=1;
	}
	else
	{
		ccwPz(0,0,0,xc,yc,zc,rr, &v[90], &u[90]); //ant face cyan
		red[6*p+5]=0;
		grn[6*p+5]=0;
		blu[6*p+5]=0;
	}

} // end cubelet0










// function to generate 12 triangles making up the faces
// of a cube centered at (xc,yc,zc) with radius=rr.
// This is 36 vertices => 108 coords
// thus the 4th parm is declared as "float v[108]"
void cubelet(
	int p,
	float xc, float yc, float zc, float rr,
	float * v, float * u )
{
	float r,g,b;

	r=red[6*p+0];
	g=grn[6*p+0];
	b=blu[6*p+0];
		ccwNx(r,g,b,xc,yc,zc,rr, &v[ 0], &u[ 0]); //lef face green


	r=red[6*p+1];
	g=grn[6*p+1];
	b=blu[6*p+1];
		ccwPx(r,g,b,xc,yc,zc,rr, &v[18], &u[18]); //rit face yellow


	r=red[6*p+2];
	g=grn[6*p+2];
	b=blu[6*p+2];
		ccwNy(r,g,b,xc,yc,zc,rr, &v[36], &u[36]); //bot face magenta

	r=red[6*p+3];
	g=grn[6*p+3];
	b=blu[6*p+3];
		ccwPy(r,g,b,xc,yc,zc,rr, &v[54], &u[54]); //top face red


	r=red[6*p+4];
	g=grn[6*p+4];
	b=blu[6*p+4];
		ccwNz(r,g,b,xc,yc,zc,rr, &v[72], &u[72]); //pos face blue

	r=red[6*p+5];
	g=grn[6*p+5];
	b=blu[6*p+5];
		ccwPz(r,g,b,xc,yc,zc,rr, &v[90], &u[90]); //ant face cyan

} // end cubelet
















// this will generate 26 cubelets centered at (0,0,0)
// with halfedge measure of rr
//
// arrays u,v need to have a dimension of 6*2*3*3*26=108*26
//
//
void rubic0(float rr, float * v, float * u)
{
//  top = red,    bot = mag(1,0,1)
//  lef = grn,    rit = yel(1,1,0)
//  ant = blu,    pos = cya(0,1,1)
	const float shrink(0.95);
	const float rrr(2.0*rr/3.0); //outer cube center @ (2/3,2/3,2/3)
	int k(0);
	for(int row=0; row<3; row++)
	{
		const float bval=0.3+float(row)/3.0f; // 0.3, 0.63, 0.96
		const float yy=(row-1)*rrr;
		for(int col=0; col<3; col++)
		{
			const float gval=0.2+float(col)/3.0f; // 0.2, 0.53, 0.86
			const float xx=(col-1)*rrr;
			for(int lay=0; lay<3; lay++)
			{
				const float rval=0.3+float(lay)/3.0f; // 0.3, 0.63, 0.96
				const float zz=(lay-1)*rrr;
				const int idx=indx(row,col,lay); // idx=0..26
				const int p=perm[idx];
				assert( p == idx ); // 1st time thru
				xxx[p]=xx;
				yyy[p]=yy; //define geometry & color for all positions
				zzz[p]=zz;

				red27[p]=rval;
				grn27[p]=gval;
				blu27[p]=bval;

				if( (row != brow) || (col != bcol) || (lay != blay) )
				{
					cubelet0(p,xx,yy,zz, shrink*rrr/2.0, &v[k], &u[k]);
					if( rgbskin )
					{
						for(int i=0; i<108; i+=3)
						{
							u[k+i+0]=red27[p];
							u[k+i+1]=grn27[p];
							u[k+i+2]=blu27[p];
						}
					}
					k += 108;
				}

			} // end for lay
		} // end for col
	} // end for row
} // end rubic0


// subsequent calls rearrange only the colors
void rubic(float rr, float * v, float * u)
{
	const float shrink(0.95);
	const float rrr(2.0*rr/3.0);
	float xx,yy,zz;
	int k(0);
	for(int row=0; row<3; row++)
	{
		for(int col=0; col<3; col++)
		{
			for(int lay=0; lay<3; lay++)
			{
				const int idx=indx(row,col,lay);
				const int p=perm[idx];
				xx=xxx[idx]; yy=yyy[idx]; zz=zzz[idx]; // fixed geometry

				if( (row != brow) || (col != bcol) || (lay != blay) )
				{
					cubelet(p,xx,yy,zz, shrink*rrr/2.0, &v[k], &u[k]);

					if( rgbskin )
					{
						for(int i=0; i<108; i+=3)
						{
							u[k+i+0]=red27[p];
							u[k+i+1]=grn27[p];
							u[k+i+2]=blu27[p];
						}
					}

					k += 108;
				} // end if

			} // end for lay
		} // end for col
	} // end for row
} // end rubic




/////////////////////////////////////////////////////











bool xtest4winner( void )
{
	bool localwin;

	localwin = (brow==1) && (bcol==1) && (blay==1);

	for(int i=0; i<size; i++)
	for(int j=0; j<size; j++) 
	for(int k=0; k<size; k++)
			localwin= (localwin && (perm[indx(i,j,k)] == indx(i,j,k)));

	return localwin;
}





int manhattan(int ra, int ca, int la, int rb, int cb, int lb) {
	return abs(ra-rb)+abs(ca-cb)+abs(la-lb);
}

// a more robust test for goal configuration;
// does not rely on "A" being in its 
// original location and orientation
// (thanks to Ryan B.)
void test4winner(void) {

bool bb,vs,vy,vu,vw,rz,tx,jr,lt,ox,qz,ks,py,mu,nw,az,cx,gt,ir;
int r[size*size*size],c[size*size*size],l[size*size*size];
int num;

// this test assumes blank in middle of cube

	for(int row=0; row<size; row++)
	for(int col=0; col<size; col++)
	for(int lay=0; lay<size; lay++) {
		num=perm[ indx(row,col,lay) ];
		r[num]=row;
		c[num]=col;
		l[num]=lay;
	}

// first check lower layer carefully:

	// v/s are 1 apart
	vs =( 1 == manhattan(r[22],c[22],l[22], r[19],c[19],l[19]) );

	// v/y are 1 apart
	vy =( 1 == manhattan(r[22],c[22],l[22], r[24],c[25],l[25]) );

	// v/u are 1 apart
	vu =( 1 == manhattan(r[22],c[22],l[22], r[21],c[21],l[21]) );

	// v/w are 1 apart
	vw =( 1 == manhattan(r[22],c[22],l[22], r[23],c[23],l[23]) );


	// r/z are 4 apart
	rz =( 4 == manhattan(r[18],c[18],l[18], r[26],c[26],l[26]) );

	// t/x are 4 apart
	tx =( 4 == manhattan(r[20],c[20],l[20], r[24],c[24],l[24]) );



// now compare lower/middle layers:

	// j/r are 1 apart
	jr =( 1 == manhattan(r[9],c[9],l[9], r[18],c[18],l[18]) );

	// l/t are 1 apart
	lt =( 1 == manhattan(r[11],c[11],l[11], r[20],c[20],l[20]) );

	// o/x are 1 apart
	ox =( 1 == manhattan(r[15],c[15],l[15], r[24],c[24],l[24]) );

	// q/z are 1 apart
	qz =( 1 == manhattan(r[17],c[17],l[17], r[26],c[26],l[26]) );

	// k/s are 1 apart
	ks =( 1 == manhattan(r[10],c[10],l[10], r[19],c[19],l[19]) );

	// p/y are 1 apart
	py =( 1 == manhattan(r[16],c[16],l[16], r[25],c[25],l[25]) );

	// m/u are 1 apart
	mu =( 1 == manhattan(r[12],c[12],l[12], r[21],c[21],l[21]) );

	// n/w are 1 apart
	nw =( 1 == manhattan(r[14],c[14],l[14], r[23],c[23],l[23]) );



// now compare top/lower layer:

	// a/z are 6 apart
	az =( 6 == manhattan(r[0],c[0],l[0], r[26],c[26],l[26]) );

	// c/x are 6 apart
	cx =( 6 == manhattan(r[2],c[2],l[2], r[24],c[24],l[24]) );

	// g/t are 6 apart
	gt =( 6 == manhattan(r[6],c[6],l[6], r[20],c[20],l[20]) );

	// i/r are 6 apart
	ir =( 6 == manhattan(r[8],c[8],l[8], r[18],c[18],l[18]) );


	// blank space is where it belongs:
	bb = (brow==1) && (bcol==1) && (blay==1);
	// note:  these indices are always correct


// now, all these conditions should be enough:

	winner = 
		vs && vy && vu && vw && rz && tx && 
		jr && lt && ox && qz && ks && py && mu && nw &&
		az && cx && gt && ir && bb;

	if( winner ) cout<<"### Winner ###"<<endl;

	bool oldwin=xtest4winner();
	if( !winner && oldwin )
	{
		cout<<"ALERT : old test4win=true but new=false"<<endl;
	}
	else if( winner && !oldwin )
	{
		cout<<"ALERT : old test4win=false but new=true"<<endl;
	}

} // end test4winner























void movenear(void)
{
	if(blay<size-1)
	{
		perm[indx(brow,bcol,blay)] = perm[indx(brow,bcol,blay+1)];
		blay += 1;
		perm[indx(brow,bcol,blay)] = 13;
	}
}



void moveaway(void)
{
	if(blay>0)
	{
		perm[indx(brow,bcol,blay)] = perm[indx(brow,bcol,blay-1)];
		blay -= 1;
		perm[indx(brow,bcol,blay)] = 13;
	}
}





void moveleft(void)
{
	if(bcol<size-1)
	{
		perm[indx(brow,bcol,blay)] = perm[indx(brow,bcol+1,blay)];
		bcol += 1;
		perm[indx(brow,bcol,blay)] = 13;
	}
}

void moveright(void)
{
	if(bcol>0)
	{
		perm[indx(brow,bcol,blay)] = perm[indx(brow,bcol-1,blay)];
		bcol -= 1;
		perm[indx(brow,bcol,blay)] = 13;
	}
}










void moveup(void)
{
	if(brow<size-1)
	{
		perm[indx(brow,bcol,blay)] = perm[indx(brow+1,bcol,blay)];
		brow += 1;
		perm[indx(brow,bcol,blay)] = 13;
	}
}

void movedown(void)
{
	if(brow>0)
	{
		perm[indx(brow,bcol,blay)] = perm[indx(brow-1,bcol,blay)];
		brow -= 1;
		perm[indx(brow,bcol,blay)] = 13;
	}
}
















void shuffle( int level )
{
	int n;

	show_goal=false;
	for(int row=0; row<3; row++)
	for(int col=0; col<3; col++)
	for(int lay=0; lay<3; lay++)
	{
		perm[ indx(row,col,lay) ] = indx(row,col,lay);
	} // end 3 loops
	brow=1;
	bcol=1;
	blay=1;


	switch( level ) {
		case 0:
			n=2;
			break;
		case 1:
			n=10;
			break;
		case 2:
			n=100;
			break;
		case 3:
			n=1000;
			break;
		case 4:
			n=10000;
			break;
		case 5:
			n=100000;
			break;
		default:
			n=0;
			break;
	} // end switch

	cout<<"shuffle, n="<<n<<endl;

// autosolve...uses stack of letters to store soln
// warning:  for large n, soln if very far from optimal.
// logic:  if blank actually moves, and if move did
//		not undo previous move, then save.

	int r1,r2,c1,c2,l1,l2;
	nsol=0;
	char prev('z');

	for(int i=0; i<=n; i++)
	{
		double r=unif01(rng);
		r1=brow; c1=bcol; l1=blay;

		if     ( (r<0.16) && (prev!='r') ) 
		{
			moveleft(); //attempted move
			r2=brow; c2=bcol; l2=blay;
			if( (r2!=r1) || (c2!=c1) || (l2!=l1) )
			{
				prev='l';
				nsol++; sol[nsol]='r';
			}
		}

		else if( (r<0.33)  && (prev!='l') ) 
		{
			moveright(); //attempted move
			r2=brow; c2=bcol; l2=blay;
			if( (r2!=r1) || (c2!=c1) || (l2!=l1) )
			{
				prev='r';
				nsol++; sol[nsol]='l';
			}
		}

		else if( (r<0.49)  && (prev!='d') ) 
		{
			moveup();    //attempted move
			r2=brow; c2=bcol; l2=blay;
			if( (r2!=r1) || (c2!=c1) || (l2!=l1) )
			{
				prev='u';
				nsol++; sol[nsol]='d';
			}
		}

		else if( (r<0.66)  && (prev!='u') ) 
		{
			movedown();  //attempted move
			r2=brow; c2=bcol; l2=blay;
			if( (r2!=r1) || (c2!=c1) || (l2!=l1) )
			{
				prev='d';
				nsol++; sol[nsol]='u';
			}
		}

		else if( (r<0.82)  && (prev!='a') ) 
		{
			movenear();  //attempted move
			r2=brow; c2=bcol; l2=blay;
			if( (r2!=r1) || (c2!=c1) || (l2!=l1) )
			{
				prev='n';
				nsol++; sol[nsol]='a';
			}
		}

		else               if (prev!='n') 
		{
			moveaway();  //attempted move
			r2=brow; c2=bcol; l2=blay;
			if( (r2!=r1) || (c2!=c1) || (l2!=l1) )
			{
				prev='a';
				nsol++; sol[nsol]='n';
			}
		}

	}

	winner=false;

} // end shuffle












void handle_mouse_dragLeft( void )
{
	int mousedx, mousedy;
	SDL_GetRelativeMouseState(&mousedx, &mousedy);

	if ( abs(mousedy) > 50 ) mousedy=0; 
	// avoid trouble getting started!
	if ( abs(mousedx) > 50 ) mousedx=0;

	float xrot = mousedy*0.25;
	float yrot = mousedx*0.25;

	glm::vec3 xaxis(1.0f,0.0f,0.0f);
	glm::vec3 yaxis(0.0f,1.0f,0.0f);
	ModelMatrix = glm::mat4(1.0);
	ModelMatrix = glm::rotate( ModelMatrix, xrot, xaxis );
	ModelMatrix = glm::rotate( ModelMatrix, yrot, yaxis );
	ModelMatrix = ModelMatrix * cumModelMatrix; //apply prior rotations
	cumModelMatrix = ModelMatrix; // update cumulative for next time

} // end handle_mouse_dragLeft




static double prevTime(0);
void handle_mouse_clickRight( float mx, float my, float Wwid, float Whit, double nowTime )
{

//cout<<"Mouse @ "<<mx<<":"<<my<<",  window : "<<Wwid<<":"<<Whit<<endl;

	if( (nowTime-prevTime) > 0.5 ) {

		prevTime = nowTime;

		int xx(mx), yy(Whit-my);
		float fmx( xx/float(Wwid) );
		float fmy( yy/float(Whit) );
		float fmz( 0.0 );

		winner=false;

		int srow,scol,slay;

// Perform a geometric pick...
// find the row,col,lay of cubelet center whose projection is closest
// to mouse position

		MVP = ProjectionMatrix * ViewMatrix * ModelMatrix;

			float mindist(1000.0);
			float xndc,yndc,zndc, xn,yn,zn, dist;
			for(int row=0; row<3; row++)
				for(int col=0; col<3; col++)
					for(int lay=0; lay<3; lay++)
						if( (row!=brow)|| (col!=bcol) || (lay!=blay) )
						{
							int i=indx(row,col,lay);
							int p=perm[ i ];
							glm::vec4 v4 = glm::vec4( xxx[i], yyy[i], zzz[i], 1.0f );
							glm::vec4 vcc = MVP * v4;
							xndc=vcc.x/vcc.w; // ndc in [-1..1]
							yndc=vcc.y/vcc.w; // 
							zndc=vcc.z/vcc.w; // bigger zndc if further away
							xn=(1+xndc)/2.0;
							yn=(1+yndc)/2.0;
							zn=(1+zndc)/2.0;
							dist = (fmx-xn)*(fmx-xn)+(fmy-yn)*(fmy-yn)+(fmz-zn)*(fmz-zn);
							if( dist < mindist )
							{
								mindist=dist;
								srow=row; scol=col; slay=lay;
							}
						}

			int dif = abs(srow-brow) + abs(scol-bcol) + abs(slay-blay);


			if( dif == 1 )
			{
				if( srow != brow )
				{
					if(srow>brow) moveup();
					else         movedown();
				}
				else if( scol != bcol )
				{
					if(scol>bcol) moveleft();
					else         moveright();
				}
				else if( slay != blay )
				{
					if(slay>blay) movenear();
					else         moveaway();
				}

				test4winner();

			} // end if diff == 1

	} // end if time delta Ok

} // end handle_mouse_clickRight












void InitSDL(Uint32 width, Uint32 height, Uint32 flags, string name)
{


	int error;
	error = SDL_Init(SDL_INIT_EVERYTHING);
	assert( error == 0 );

	// this is essential to avoid SDL, GL error messages
	assert( SDL_GL_LoadLibrary( nullptr ) >= 0 );
	//assert( SDL_GL_LoadLibrary( "/usr/lib64/libGL.so" ) >= 0 );


	//make buffer swap synchronized w/vertical refresh:
	SDL_GL_SetSwapInterval(1);

	assert( SDL_SetHint( SDL_HINT_RENDER_VSYNC, "1" ) );
	assert( SDL_SetHint( SDL_HINT_RENDER_SCALE_QUALITY, "1" ) );


	///////////////////////////////////////////////////////////////////////

	// WARNING:  don't know how to print valid numbers using cout
	SDL_version compiled;
	SDL_VERSION(&compiled);
	printf("We compiled against SDL version %d.%d.%d ...\n",
			compiled.major, compiled.minor, compiled.patch);

	// WARNING:  don't know how to print valid numbers using cout
	SDL_version linked;
	SDL_GetVersion(&linked);
	printf("We linked against SDL version %d.%d.%d.\n",
			linked.major, linked.minor, linked.patch);


	///////////////////////////////////////////////////////////////////////


	// Turn on double buffering.
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 24);
	SDL_GL_SetAttribute(SDL_GL_ALPHA_SIZE, 8);//for alpha<1




////////////// begin try ///////////////////////////////////////////

	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, 3); //
	SDL_GL_SetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, 3); //
	SDL_GL_SetAttribute( SDL_GL_CONTEXT_PROFILE_MASK, SDL_GL_CONTEXT_PROFILE_CORE );

	int major,minor;
	SDL_GL_GetAttribute(SDL_GL_CONTEXT_MAJOR_VERSION, &major); //
	SDL_GL_GetAttribute(SDL_GL_CONTEXT_MINOR_VERSION, &minor); //
	cout<<"SDL query: OpenGL version major="<<major<<", minor="<<minor<<endl;


	///////////////////////////////////////////////////////////////////////




	// Create the window
	mainWindow = SDL_CreateWindow(name.c_str(), 
			SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, 
			width, height, flags);
	if( mainWindow == nullptr )
	{
	  cout<<"SDL Error @ CreateWindow: "<<SDL_GetError()<<endl;
	  exit(-1);
	}


	mainGLContext = SDL_GL_CreateContext(mainWindow);
	if( mainGLContext == nullptr )
	{
	  cout<<"SDL Error @ CreateContext: "<<SDL_GetError()<<endl;
	  cout<<"@ line "<<__LINE__<<endl;
	  exit(-1);
	}


	SDL_GL_MakeCurrent( mainWindow, mainGLContext );



} // end InitSDL






// Initial position on +Z
float camradius(5.0);
// Initial horizontal angle : toward -Z
static float horiAng = 3.14f;
// Initial vertical angle : none
static float vertAng = 0.0f;
// Initial Field of View
static float fovdeg = 45.0f;




void updateVP( int width, int height ) {

	// Direction : Spherical coordinates to Cartesian coordinates conversion
	glm::vec3 direction(
		cos(vertAng) * sin(horiAng), 
		sin(vertAng),
		cos(vertAng) * cos(horiAng)
	);


	// Right vector
	glm::vec3 right = glm::vec3(sin(horiAng - halfpi), 0, cos(horiAng - halfpi));
	
	// Up vector
	glm::vec3 up = glm::cross( right, direction );

	glm::vec3 position = glm::vec3( -camradius*direction ); 



	// careful:  near cutoff must be < margin
	float const near(0.1);
	float const far(100.0);
	ProjectionMatrix = glm::perspective(fovdeg, float(width)/float(height), near, far);
	ViewMatrix       = glm::lookAt( position,	position+direction, up );


} // end updateMVP










////////////////////////////////////////////////////
// main program 

int main( int argc,  char * argv[]  )
{
	if( !sbClap.loadFromFile("data/applause_short.wav") ) return -1;


	if( argc > 1 ) // at least 1 command line parm given
	{
		rgbskin=true; // rgb cubelet colors
		cout<<" command line parm: "<<argv[1]<<endl;
	}
	else // no command line parms were given
	{
		rgbskin=false; // normal rufas cube colors
	}



	double playSecs;
	bool playedonce(false);


	int Wsize(500);
	int Wwid(Wsize), Whit(Wsize); //initial window size


	for(int i=0; i<27; i++) perm[i]=i;




	Uint32 contextFlags;
	contextFlags = 
	SDL_WINDOW_SHOWN
	|SDL_WINDOW_OPENGL
	|SDL_WINDOW_ALLOW_HIGHDPI
	|SDL_WINDOW_RESIZABLE;
	// new version for Mac Retina high dpi

	InitSDL(Wwid, Whit, contextFlags, "RufasCube");




	// Initialize GLEW
	glewExperimental = true; // Needed for core profile
	if (glewInit() != GLEW_OK) {
		fprintf(stderr, "Failed to initialize GLEW\n");
		return -1;
	}



	int Fwid, Fhit;
	SDL_GL_GetDrawableSize( mainWindow, &Fwid, &Fhit );
	cout<<"Drawable size: "<<Fwid<<"-by-"<<Fhit<<endl<<endl; 
	// 2880-by-1800 for Retina (which is the double the screensize)...
	// These values must now be used in glViewport call, rather than Wwid, Whit.

	glViewport(0,0,Fwid,Fhit);




	//This is the new way for getting the GL version.
	//Much better than the old glGetString(GL_VERSION).
	int OpenGLVersion[2];
	glGetIntegerv(GL_MAJOR_VERSION, &OpenGLVersion[0]);
	glGetIntegerv(GL_MINOR_VERSION, &OpenGLVersion[1]);
	cout<<"OpenGL major version = "<<OpenGLVersion[0]<<endl;
	cout<<"OpenGL minor version = "<<OpenGLVersion[1]<<endl<<endl;




	// additions intended to reduce aliasing:
	glEnable(GL_MULTISAMPLE);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);






	// gray background
	glClearColor(0.2f, 0.2f, 0.2f, 1.0f);

	// Enable depth test
	glEnable(GL_DEPTH_TEST);
	// Accept fragment if it closer to the camera than the former one
	glDepthFunc(GL_LESS); 

	// Cull triangles which normal is not towards the camera
	glEnable(GL_CULL_FACE);

	GLuint VertexArrayID;
	glGenVertexArrays(1, &VertexArrayID); //create 1 vertex array
	glBindVertexArray(VertexArrayID);//bind;  subsequent VA ftns modify this one

	bool Ok;

	// Create and compile our GLSL program from the shaders
	GLuint programID = LoadShaders( "data/vertshader330.glsl", "data/fragshader330.glsl", Ok );

	// Get a handle for our "MVP" uniform
	GLuint MatrixID = glGetUniformLocation(programID, "MVP");




	// 1st call sets geometry and colors
	rubic0(1.0f, &vb[0], &rgb[0]);


	GLuint rgbbuffer;


	GLuint vertexbuffer;


	initModel(); // rotate into preferred orientation

	initText2D( "data/rods3w.png", Wwid,Whit, Ok ); //favorite



	int numkeys;
	const Uint8 * key_map = SDL_GetKeyboardState(&numkeys); //do NOT delete;  "SDL-internal" pointer

	bool userexit(false);

	double previousTime = double( SDL_GetTicks() ) / 1000.0;



	do{



///////////////////// begin response to keys /////////////////////////

		double currentTime = double( SDL_GetTicks() ) / 1000.0;

		SDL_PumpEvents();

		// some, or all, of these key events need a minimum delay:
		//double const keydlay(0.2);
		double const keydlay(0.1);

		if( fabs(currentTime - previousTime) > keydlay ) {

			previousTime=currentTime;

			if(      key_map[ SDL_SCANCODE_ESCAPE ] )	 userexit=true;
			else if( key_map[ SDL_SCANCODE_UP     ] )  moveleft();
			else if( key_map[ SDL_SCANCODE_DOWN   ] )  moveright();
			else if( key_map[ SDL_SCANCODE_LEFT   ] )  movenear();
			else if( key_map[ SDL_SCANCODE_RIGHT   ] ) moveaway();
			else if( key_map[ SDLK_PLUS   ] )  			movedown();
			else if( key_map[ SDL_SCANCODE_MINUS   ] ) moveup();

			else if( key_map[ SDL_SCANCODE_F1 ] )  shuffle(1);
			else if( key_map[ SDL_SCANCODE_F2 ] )  shuffle(2);
			else if( key_map[ SDL_SCANCODE_F3 ] )  shuffle(3);
			else if( key_map[ SDL_SCANCODE_F4 ] )  shuffle(4);
			else if( key_map[ SDL_SCANCODE_F5 ] )  shuffle(5);

			else if( key_map[ SDL_SCANCODE_F6 ] )  help=!help;

			else if( key_map[ SDL_SCANCODE_A ] )  show_axes=!show_axes;

			else if( key_map[ SDL_SCANCODE_SPACE ] )  hint=!hint;

			else if( key_map[ SDL_SCANCODE_S ] )
			{

				// solve one step at a time
				if( nsol>0 )
				{
					cout<<"Steps to solution: "<<nsol<<endl;
					char s=sol[nsol]; nsol--;
					if     ( s=='l' ) moveleft();
					else if( s=='r' ) moveright();
					else if( s=='u' ) moveup();
					else if( s=='d' ) movedown();
					else if( s=='n' ) movenear();
					else if( s=='a' ) moveaway();
					else
					{
						cout<<"illegal letter stored in sol array"<<endl;
					}
				}
				test4winner();

			}

		}



	///////////////////// begin response to mouse /////////////////////////////////////////////

		int mx,my;
		Uint32 MouseState=SDL_GetMouseState(&mx,&my);

		if( MouseState & SDL_BUTTON(1) ) 
			handle_mouse_dragLeft( );
		else if( MouseState & SDL_BUTTON(3) ) 
			handle_mouse_clickRight(mx,my,Wwid,Whit,currentTime);

		// 12jun14 added for laptop users without mouse:
		else if( key_map[ SDL_SCANCODE_RETURN ] ) 
			handle_mouse_clickRight(mx,my,Wwid,Whit,currentTime);



///////////////////// end response to key/mouse /////////////////////////////////////////////


		updateVP(Wwid,Whit);
		MVP = ProjectionMatrix * ViewMatrix * ModelMatrix;



		//SDL_GetWindowSize( mainWindow, &Wwid, &Whit );
		//SDL_GL_GetDrawableSize( mainWindow, &Fwid, &Fhit );
		//glViewport(0,0,Fwid,Fhit);

		int Nwid,Nhit;
		SDL_GetWindowSize( mainWindow, &Nwid, &Nhit );
		if( (Nwid != Wwid) || (Nhit != Whit) )
		{
			Wwid=Nwid;  Whit=Nhit;
			resizeText2D(Wwid,Whit);
			SDL_GL_GetDrawableSize( mainWindow, &Fwid, &Fhit );
			glViewport(0,0,Fwid,Fhit);
		}












		// subsequent calls permute cubes & colors
		rubic( 1.0f, &vb[0], &rgb[0] );

		glGenBuffers(1, &rgbbuffer);
		glBindBuffer(GL_ARRAY_BUFFER, rgbbuffer);
		glBufferData(GL_ARRAY_BUFFER, sizeof(rgb), rgb, GL_STATIC_DRAW);

		glGenBuffers(1, &vertexbuffer);
		glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
		glBufferData(GL_ARRAY_BUFFER, sizeof(vb), vb, GL_STATIC_DRAW);

/*
		// keep it square!
		glfwGetWindowSize( mainWin, &Wwid, &Whit );
		Wsize = (Wwid+Whit)/2;  Wwid=Wsize; Whit=Wsize;
		glViewport(0,0,Wwid,Whit);
		glfwSetWindowSize( mainWin, Wwid, Whit );
*/

		// Clear the screen
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		// Use our shader
		glUseProgram(programID);


		// Send our transformation to the currently bound shader, 
		// in the "MVP" uniform
		glUniformMatrix4fv(MatrixID, 1, GL_FALSE, &MVP[0][0]);

		// 1st attribute buffer : vertices
		glEnableVertexAttribArray(0);
		glBindBuffer(GL_ARRAY_BUFFER, vertexbuffer);
		glVertexAttribPointer(0,3,	GL_FLOAT,GL_FALSE, 0, (void*)0 );

		// 2nd attribute buffer : UVs
		glEnableVertexAttribArray(1);
		glBindBuffer(GL_ARRAY_BUFFER, rgbbuffer);
		glVertexAttribPointer(1, 3, GL_FLOAT, GL_TRUE, 0, (void*)0 );

		// Draw the triangles !
		glDrawArrays(GL_TRIANGLES, 0, 26*12*3); // 26cubes => 26*12 triangles


		glDisableVertexAttribArray(0);
		glDisableVertexAttribArray(1);




		//printText2d("LL", 0.05f, 0.05f, 15);
		//printText2d("C", 0.5f, 0.5f, 15);


		///// fixed position text messages:

		printText2d("<esc>=exit       <spc>=alpha", 0.05f, 0.95f, 15);
		printText2d("<F1:F5>=shuffle    <F6>=help", 0.05f, 0.02f, 15);


		if( help )
		{
			printText2d("3D Block Slider Puzzle...",         0.02f, 0.85f, 10);
			printText2d("Restore initial order of cubes",    0.02f, 0.80f, 10);
			printText2d("based on color or alphabet;",       0.02f, 0.75f, 10);
			printText2d("Note empty space in center.",       0.02f, 0.70f, 10);
			printText2d("Ftn keys shuffle cubelets;",        0.02f, 0.65f, 10);
			printText2d("Left mouse btn rotates puzzle;",    0.02f, 0.60f, 10);
			printText2d("Right mouse btn picks cubelet",     0.02f, 0.55f, 10);
			printText2d("to slide into empty space.",        0.02f, 0.50f, 10);

			printText2d("          Keys work too:",          0.02f, 0.25f, 10);
			printText2d("<PgUp>= -Z  <Up>= +Y  <Rt>= +X",    0.02f, 0.20f, 10);
			printText2d("<PgDn>= +Z  <Dn>= -Y  <Lt>= -X",    0.02f, 0.15f, 10);
			printText2d("            <a>=axes",              0.02f, 0.10f, 10);

		}


		////// moving text


		if( hint )
		{

			// begin find median Z in normal.dev.coords
			int j=indx(1,1,1); //center cubelet
			int q=perm[j];
			glm::vec4 w4 = glm::vec4( xxx[j],yyy[j],zzz[j], 1.0f );
			glm::vec4 wcc = MVP * w4;
			float median_zndc=wcc.z/wcc.w; // use 1.0 when not available
			// this helps readability by labelling only the closest cubelets
			// end median Z

			assert( symbol[ perm[ indx(brow,bcol,blay) ] ] == " " );
			for(int row=0; row<3; row++)
				for(int col=0; col<3; col++)
					for(int lay=0; lay<3; lay++)
						if( (row!=brow)|| (col!=bcol) || (lay!=blay) )
						{
							int i=indx(row,col,lay);
							int p=perm[ i ];
							glm::vec4 v4 = glm::vec4( xxx[i], yyy[i], zzz[i], 1.0f );
							glm::vec4 vcc = MVP * v4;
							printText3D( 
									symbol[ p ].c_str(), 
									vcc.x, vcc.y, vcc.z, vcc.w,
									30, median_zndc );
						}

////////////////////////////////////////////////////////



		} // end if hint


////////////////////////////////////////////////////////

		if (show_axes)
		{
			// show X, Y, Z coordinate axes:
			glm::vec4 v4 = glm::vec4( 1.1, -1.1, -1.1, 1.0f );
			glm::vec4 vcc = MVP * v4;
			printText3D("+X",vcc.x,vcc.y,vcc.z,vcc.w,10,1.0);

			v4 = glm::vec4( -1.1, 1.1, -1.1, 1.0f );
			vcc = MVP * v4;
			printText3D("+Y",vcc.x,vcc.y,vcc.z,vcc.w,10,1.0);

			v4 = glm::vec4( -1.1, -1.1, 1.1, 1.0f );
			vcc = MVP * v4;
			printText3D("+Z",vcc.x,vcc.y,vcc.z,vcc.w,10,1.0);
		}

////////////////////////////////////////////////////////


		if( winner )
		{
			printText2d("Correct !", 0.08, 0.5, 30);
		}



		glFlush();
		SDL_GL_SwapWindow( mainWindow );



		if( winner )
		{

			if( !playedonce  ) 
			{

				sound.setBuffer(sbClap);
				sound.setVolume(90);
				sound.setLoop(false);
				sound.play();

				playedonce=true;
				playSecs = SDL_GetTicks() / 1000.0;
			}

			double elapsedSec = SDL_GetTicks()/1000.0 - playSecs;
			if( elapsedSec > 8.0 )	winner=false;

		}
		else {
			playedonce=false;
		}





	} // Check if the ESC key was pressed or the window was closed
	while( !userexit );

	// Cleanup VBO and shader
	glDeleteBuffers(1, &vertexbuffer);
	glDeleteBuffers(1, &rgbbuffer);
	glDeleteProgram(programID);
	glDeleteVertexArrays(1, &VertexArrayID);


	// Delete the text's VBO, the shader and the texture
	cleanupText2D();


	SDL_GL_DeleteContext(mainGLContext);
	SDL_DestroyWindow(mainWindow);
	SDL_Quit();


	return 0;
}

