

/*
--
-- Copyright (C) 2014  <fastrgv@gmail.com>
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You may read the full text of the GNU General Public License
-- at <http://www.gnu.org/licenses/>.
--
*/




#include <vector>
#include <cstring>
#include <GL/glew.h>

#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
using namespace glm;

#include "shader.hpp"

#include "text2d.hpp"
#include "loadpng.hpp"


unsigned int Text2DTextureID;
unsigned int Text2DVertexBufferID;
unsigned int Text2DUVBufferID;
unsigned int Text2DShaderID;
unsigned int samplerUniformID;
unsigned int centerUniformID;

int Wwid, Whit;





void resizeText2D(int Swid, int Shit){

	Wwid=Swid;
	Whit=Shit;

}


void initText2D(const char * texturePath, int Swid, int Shit, bool &Ok){

	Wwid=Swid;
	Whit=Shit;

	// Initialize texture
	Text2DTextureID = loadPNG(texturePath,Ok);

	// Initialize VBO
	glGenBuffers(1, &Text2DVertexBufferID);
	glGenBuffers(1, &Text2DUVBufferID);

	// Initialize Shader
	Text2DShaderID = LoadShaders( "./data/text2d.vs", "./data/text2d.fs", Ok );

	if( Ok )
		samplerUniformID = glGetUniformLocation( Text2DShaderID, "myTextureSampler" );

	if( Ok )
		centerUniformID = glGetUniformLocation( Text2DShaderID, "screenCen" );

}




/*
void printText3D(
	const char * text, 
	float ccx, float ccy, float ccw, 
	int isize, float range
){

	float fsize( float(isize)/range );
	int nusize( fsize);

	float hfrac( 0.5 * float(nusize) / float(Wwid)  );
	float vfrac( 0.5 * float(nusize) / float(Whit)  );

	float const xoff(-hfrac), yoff(-vfrac);
	float const xfac(1.0), yfac(1.0);

	float xcen,ycen;

	float xndc,yndc; // Normalized Device coords (ndc)
	xndc = ccx/ccw; yndc=ccy/ccw; // ndc in [-1:1]
	xcen=(1.0+xndc)/2.0;  ycen=(1.0+yndc)/2.0;// in [0:1]
	printText2D(text, Wwid*(xfac*xcen+xoff), Whit*(yfac*ycen+yoff), nusize);

} // end printText3D
*/

void printText3D(
	const char * text, 
	float ccx, float ccy, float ccz, float ccw, 
	int size, 
	float Mzndc  // median zndc (bigger is further away from camera)
){

	float const xoff(-0.03), yoff(-0.03);
	float const xfac(1.0), yfac(1.0);
	float const eps(0.001f);

	float xcen,ycen;

	float xndc,yndc,zndc; // Normalized Device coords (ndc)
	xndc = ccx/ccw; yndc=ccy/ccw; zndc=ccz/ccw;// ndc in [-1:1]
	xcen=(1.0+xndc)/2.0;  ycen=(1.0+yndc)/2.0;// in [0:1]
	//if( (-1.0 < zndc) && (zndc < 1.0) ) //revert to this if median trick fails
	if( (-1.0 < zndc) && (zndc < Mzndc-eps) )
		printText2D(text, Wwid*(xfac*xcen+xoff), Whit*(yfac*ycen+yoff), size);

} // end printText3D





void printText2d(const char * text, float xcen, float ycen, int fsize){

	float hfrac( 0.5 * float(fsize) / float(Wwid)  );
	float vfrac( 0.5 * float(fsize) / float(Whit)  );

	float const xoff(-hfrac), yoff(-vfrac);
	float const xfac(1.0), yfac(1.0);

	int xpos = Wwid*(xfac*xcen+xoff);
	int ypos = Whit*(yfac*ycen+yoff);

	printText2D( text, xpos, ypos, fsize );

} // end printText2d


// original
void printText2D(const char * text, int x, int y, int fsize){

	unsigned int length = strlen(text);

	// Fill buffers
	std::vector<glm::vec2> vertices;
	std::vector<glm::vec2> UVs;
	for ( unsigned int i=0 ; i<length ; i++ ){
		
		glm::vec2 vertex_up_left    = glm::vec2( x+i*fsize      , y+fsize );
		glm::vec2 vertex_up_right   = glm::vec2( x+i*fsize+fsize, y+fsize );
		glm::vec2 vertex_down_right = glm::vec2( x+i*fsize+fsize, y      );
		glm::vec2 vertex_down_left  = glm::vec2( x+i*fsize     , y      );

		vertices.push_back(vertex_up_left   );
		vertices.push_back(vertex_down_left );
		vertices.push_back(vertex_up_right  );

		vertices.push_back(vertex_down_right);
		vertices.push_back(vertex_up_right);
		vertices.push_back(vertex_down_left);

		char character = text[i];
		float uv_x = (character%16)/16.0f;
		float uv_y = (character/16)/16.0f;

		glm::vec2 uv_up_left    = glm::vec2( uv_x           , 1.0f - uv_y );
		glm::vec2 uv_up_right   = glm::vec2( uv_x+1.0f/16.0f, 1.0f - uv_y );
		glm::vec2 uv_down_right = glm::vec2( uv_x+1.0f/16.0f, 1.0f - (uv_y + 1.0f/16.0f) );
		glm::vec2 uv_down_left  = glm::vec2( uv_x           , 1.0f - (uv_y + 1.0f/16.0f) );
		UVs.push_back(uv_up_left   );
		UVs.push_back(uv_down_left );
		UVs.push_back(uv_up_right  );

		UVs.push_back(uv_down_right);
		UVs.push_back(uv_up_right);
		UVs.push_back(uv_down_left);
	}
	glBindBuffer(GL_ARRAY_BUFFER, Text2DVertexBufferID);
	glBufferData(GL_ARRAY_BUFFER, vertices.size() * sizeof(glm::vec2), &vertices[0], GL_STATIC_DRAW);
	glBindBuffer(GL_ARRAY_BUFFER, Text2DUVBufferID);
	glBufferData(GL_ARRAY_BUFFER, UVs.size() * sizeof(glm::vec2), &UVs[0], GL_STATIC_DRAW);

	// Bind shader
	glUseProgram(Text2DShaderID);

	// Bind texture
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, Text2DTextureID);
	// Set our "myTextureSampler" sampler to user Texture Unit 0
	glUniform1i(samplerUniformID, 0);
	glUniform2i(centerUniformID, Wwid/2, Whit/2 );

	// 1rst attribute buffer : vertices
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, Text2DVertexBufferID);
	glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 0, (void*)0 );

	// 2nd attribute buffer : UVs
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, Text2DUVBufferID);
	glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, (void*)0 );

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Draw call
	glDrawArrays(GL_TRIANGLES, 0, vertices.size() );

	glDisable(GL_BLEND);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);

}

void cleanupText2D(){

	// Delete buffers
	glDeleteBuffers(1, &Text2DVertexBufferID);
	glDeleteBuffers(1, &Text2DUVBufferID);

	// Delete texture
	glDeleteTextures(1, &Text2DTextureID);

	// Delete shader
	glDeleteProgram(Text2DShaderID);
}
